@@ -29,6 +29,7 @@
 #include <sys/stat.h>
 #include <regex.h>
 #include <time.h>
+#include <sys/time.h>
 #include <unistd.h>
 #include <limits.h>
 
@@ -69,9 +70,36 @@ static struct pseudo_variables pseudo_env[] = {
 #ifdef PSEUDO_PROFILING
 	{ "PSEUDO_PROFILE_PATH", 19, NULL },
 #endif
+	{ "PSEUDO_EVLOG", 12, NULL },
+	{ "PSEUDO_EVLOG_FILE", 17, NULL },
 	{ NULL, 0, NULL } /* Magic terminator */
 };
 
+typedef struct {
+	struct timeval stamp;
+	int len;
+	char *data;
+} pseudo_evlog_entry;
+
+/* so bash overrides getenv/unsetenv/etcetera, preventing them from
+ * actually modifying environ, so we have pseudo_wrappers try to dlsym
+ * the right values. This could fail, in which case we'd get null
+ * pointers, and we'll just call whatever the linker gives us and
+ * hope for the best.
+ */
+#define SETENV(x, y, z) (pseudo_real_setenv ? pseudo_real_setenv : setenv)(x, y, z)
+#define GETENV(x) (pseudo_real_getenv ? pseudo_real_getenv : getenv)(x)
+#define UNSETENV(x) (pseudo_real_unsetenv ? pseudo_real_unsetenv : unsetenv)(x)
+
+#define PSEUDO_EVLOG_ENTRIES 250
+#define PSEUDO_EVLOG_LENGTH 256
+static pseudo_evlog_entry event_log[PSEUDO_EVLOG_ENTRIES];
+static char *pseudo_evlog_buffer;
+static int pseudo_evlog_next_entry = 0;
+static void pseudo_evlog_set(char *);
+static void pseudo_evlog_flags_finalize(void);
+static unsigned long pseudo_debug_flags_in(char *);
+
 /* -1 - init hasn't been run yet
  * 0 - init has been run
  * 1 - init is running
@@ -83,8 +111,11 @@ static struct pseudo_variables pseudo_env[] = {
 static int pseudo_util_initted = -1;  /* Not yet run */
 
 /* bypass wrapper logic on path computations */
-int (*pseudo_real_lstat)(const char *path, struct stat *buf) = lstat;
-int (*pseudo_real_fstat)(int, struct stat *buf) = fstat;
+int (*pseudo_real_lstat)(const char *path, PSEUDO_STATBUF *buf) = NULL;
+/* bash workaround */
+int (*pseudo_real_unsetenv)(const char *) = unsetenv;
+char * (*pseudo_real_getenv)(const char *) = getenv;
+int (*pseudo_real_setenv)(const char *, const char *, int) = setenv;
 
 #if 0
 static void
@@ -109,14 +140,14 @@ pseudo_has_unload(char * const *envp) {
 	size_t i = 0;
 
 	/* Is it in the caller environment? */
-	if (NULL != getenv(unload))
+	if (NULL != GETENV(unload))
 		return 1;
 
 	/* Is it in the environment cache? */
 	if (pseudo_util_initted == -1)
 		pseudo_init_util();
 	while (pseudo_env[i].key && strcmp(pseudo_env[i].key, unload))
-	       ++i;
+		++i;
 	if (pseudo_env[i].key && pseudo_env[i].value)
 		return 1;
 
@@ -144,7 +175,7 @@ pseudo_get_value(const char *key) {
 	/* Check if the environment has it and we don't ...
 	 * if so, something went wrong... so we'll attempt to recover
 	 */
-	if (pseudo_env[i].key && !pseudo_env[i].value && getenv(pseudo_env[i].key))
+	if (pseudo_env[i].key && !pseudo_env[i].value && GETENV(pseudo_env[i].key))
 		pseudo_init_util();
 
 	if (pseudo_env[i].value)
@@ -198,17 +229,17 @@ pseudo_init_util(void) {
 	pseudo_util_initted = 1;
 
 	for (i = 0; pseudo_env[i].key; i++) {
-		if (getenv(pseudo_env[i].key))
-			pseudo_set_value(pseudo_env[i].key, getenv(pseudo_env[i].key));
+		if (GETENV(pseudo_env[i].key))
+			pseudo_set_value(pseudo_env[i].key, GETENV(pseudo_env[i].key));
 	}
 
 	pseudo_util_initted = 0;
 
 	/* Somewhere we have to set the debug level.. */
 	env = pseudo_get_value("PSEUDO_DEBUG");
-        if (env) {
+	if (env) {
 		int i;
-                int level = atoi(env);
+		int level = atoi(env);
 		if (level > 0) {
 			for (i = 0; i < level; ++i) {
 				pseudo_debug_verbose();
@@ -217,12 +248,20 @@ pseudo_init_util(void) {
 			pseudo_debug_set(env);
 		}
 		pseudo_debug_flags_finalize();
-        }
-        free(env);
+	}
+	free(env);
+	env = pseudo_get_value("PSEUDO_EVLOG");
+	if (env) {
+		pseudo_evlog_set(env);
+		pseudo_evlog_flags_finalize();
+	}
+	free(env);
 }
 
 unsigned long pseudo_util_debug_flags = 0;
+unsigned long pseudo_util_evlog_flags = 0;
 int pseudo_util_debug_fd = 2;
+int pseudo_util_evlog_fd = 2;
 static int debugged_newline = 1;
 static char pid_text[32];
 static size_t pid_len;
@@ -390,30 +429,52 @@ pseudo_debug_verbose(void) {
 	}
 }
 
+void
+pseudo_debug_set(char *s) {
+	pseudo_util_debug_flags = pseudo_debug_flags_in(s);
+}
+
+static void
+pseudo_evlog_set(char *s) {
+	pseudo_util_evlog_flags = pseudo_debug_flags_in(s);
+}
+
 /* This exists because we don't want to allocate a bunch of strings
  * and free them immediately if you have several flags set.
  */
-void
-pseudo_debug_flags_finalize(void) {
+static void
+pseudo_flags_finalize(unsigned long flags, char *value) {
 	char buf[PDBG_MAX + 1] = "", *s = buf;
 	for (int i = 0; i < PDBG_MAX; ++i) {
-		if (pseudo_util_debug_flags & (1 << i)) {
+		if (flags & (1 << i)) {
 			*s++ = pseudo_debug_type_symbolic(i);
 		}
 	}
-	pseudo_set_value("PSEUDO_DEBUG", buf);
+	pseudo_set_value(value, buf);
 }
 
 void
-pseudo_debug_set(char *s) {
+pseudo_debug_flags_finalize(void) {
+	pseudo_flags_finalize(pseudo_util_debug_flags, "PSEUDO_DEBUG");
+}
+
+void
+pseudo_evlog_flags_finalize(void) {
+	pseudo_flags_finalize(pseudo_util_evlog_flags, "PSEUDO_EVLOG");
+}
+
+static unsigned long
+pseudo_debug_flags_in(char *s) {
+	unsigned long flags = 0;
 	if (!s)
-		return;
+		return flags;
 	for (; *s; ++s) {
 		int id = pseudo_debug_type_symbolic_id(*s);
 		if (id > 0) {
-			pseudo_util_debug_flags |= (1 << id);
+			flags |= (1 << id);
 		}
 	}
+	return flags;
 }
 
 void
@@ -455,12 +516,100 @@ pseudo_diag(char *fmt, ...) {
 	return wrote;
 }
 
+void
+pseudo_evlog_dump(void) {
+	char scratch[256], firstdate[64], lastdate[64];
+	time_t first = 0, last = 0;
+	int len;
+	int entries = 0;
+	struct tm first_tm, last_tm;
+	int wrote; /* ignoring write errors because there's nothing we can do */
+
+	for (int i = 0; i < PSEUDO_EVLOG_ENTRIES; ++i) {
+		pseudo_evlog_entry *e = &event_log[i];
+		if (!e->data || e->len < 0 || e->stamp.tv_sec == 0)
+			continue;
+		++entries;
+		if (!first || e->stamp.tv_sec < first)
+			first = e->stamp.tv_sec;
+		if (!last || e->stamp.tv_sec > last)
+			last = e->stamp.tv_sec;
+	}
+	localtime_r(&first, &first_tm);
+	localtime_r(&last, &last_tm);
+	strftime(firstdate, 64, "%Y-%M-%D %H:%M:%S", &first_tm);
+	strftime(lastdate, 64, "%Y-%M-%D %H:%M:%S", &last_tm);
+
+	len = snprintf(scratch, 256, "event log for pid %d [%d entries]\n",
+		getpid(), entries);
+	if (len > 256)
+		len = 256;
+	wrote = write(pseudo_util_evlog_fd, scratch, len);
+	len = snprintf(scratch, 256, "  first entry %s\n", firstdate);
+	wrote = write(pseudo_util_evlog_fd, scratch, len);
+	len = snprintf(scratch, 256, "  last entry %s\n", lastdate);
+	wrote = write(pseudo_util_evlog_fd, scratch, len);
+
+	for (int i = 0; i < PSEUDO_EVLOG_ENTRIES; ++i) {
+		int entry = (pseudo_evlog_next_entry + i) % PSEUDO_EVLOG_ENTRIES;
+		pseudo_evlog_entry *ev = &event_log[entry];
+		if (!ev->data || ev->len <= 0)
+			continue;
+		localtime_r(&ev->stamp.tv_sec, &first_tm);
+		len = strftime(firstdate, 64, "%H:%M:%S", &first_tm);
+		if (len) {
+			len = snprintf(scratch, 256, "%s.%03d: ", firstdate,
+				(int) (ev->stamp.tv_usec / 1000));
+			wrote = write(pseudo_util_evlog_fd, scratch, len);
+		} else {
+			wrote = write(pseudo_util_evlog_fd, "no timestamp: ", 14);
+		}
+		wrote = write(pseudo_util_evlog_fd, ev->data, ev->len);
+	}
+	(void) wrote;
+}
+
+int
+pseudo_evlog_internal(char *fmt, ...) {
+	va_list ap;
+	pseudo_evlog_entry *ev = &event_log[pseudo_evlog_next_entry++];
+
+	pseudo_evlog_next_entry %= PSEUDO_EVLOG_ENTRIES;
+
+	if (!ev->data) {
+		pseudo_evlog_buffer = malloc(PSEUDO_EVLOG_ENTRIES * PSEUDO_EVLOG_LENGTH);
+		if (pseudo_evlog_buffer) {
+			for (int i = 0; i < PSEUDO_EVLOG_ENTRIES; ++i) {
+				event_log[i].data = pseudo_evlog_buffer + (PSEUDO_EVLOG_LENGTH * i);
+			}
+		} else {
+			pseudo_diag("fatal: can't allocate event log storage.\n");
+		}
+	}
+
+	va_start(ap, fmt);
+	ev->len = vsnprintf(ev->data, PSEUDO_EVLOG_LENGTH, fmt, ap);
+	va_end(ap);
+	if (ev->len > PSEUDO_EVLOG_LENGTH) {
+		strcpy(ev->data + PSEUDO_EVLOG_LENGTH - 5, "...\n");
+		ev->len = PSEUDO_EVLOG_LENGTH - 1;
+	}
+	gettimeofday(&ev->stamp, NULL);
+
+	return ev->len;
+}
+
 /* store pid in text form for prepending to messages */
 void
 pseudo_new_pid() {
+#if PSEUDO_PORT_LINUX
+	extern char *program_invocation_short_name; /* glibcism */
+#else
+	char *program_invocation_short_name = "unknown";
+#endif
 	int pid = getpid();
 	pid_len = snprintf(pid_text, 32, "%d: ", pid);
-	pseudo_debug(PDBGF_PID, "new pid: %d\n", pid);
+	pseudo_debug(PDBGF_PID, "new pid: %d [%s]\n", pid, program_invocation_short_name);
 }
 
 /* helper function for pseudo_fix_path
@@ -473,7 +622,7 @@ pseudo_append_element(char *newpath, char *root, size_t allocated, char **pcurre
 	static int link_recursion = 0;
 	size_t curlen;
 	char *current;
-	struct stat buf;
+	PSEUDO_STATBUF buf;
 	if (!newpath ||
 	    !pcurrent || !*pcurrent ||
 	    !root || !element) {
@@ -516,7 +665,7 @@ pseudo_append_element(char *newpath, char *root, size_t allocated, char **pcurre
 	/* if lstat fails, that's fine -- nonexistent files aren't symlinks */
 	if (!leave_this) {
 		int is_link;
-		is_link = (pseudo_real_lstat(newpath, &buf) != -1) && S_ISLNK(buf.st_mode);
+		is_link = (pseudo_real_lstat) && (pseudo_real_lstat(newpath, &buf) != -1) && S_ISLNK(buf.st_mode);
 		if (link_recursion >= PSEUDO_MAX_LINK_RECURSION && is_link) {
 			pseudo_diag("link recursion too deep, not expanding path '%s'.\n", newpath);
 			is_link = 0;
@@ -685,7 +834,7 @@ pseudo_fix_path(const char *base, const char *path, size_t rootlen, size_t basel
  * we don't try to fix the library path.
  */
 void pseudo_dropenv() {
-	char *ld_preload = getenv(PRELINK_LIBRARIES);
+	char *ld_preload = GETENV(PRELINK_LIBRARIES);
 
 	if (ld_preload) {
 		ld_preload = without_libpseudo(ld_preload);
@@ -693,10 +842,9 @@ void pseudo_dropenv() {
 			pseudo_diag("fatal: can't allocate new %s variable.\n", PRELINK_LIBRARIES);
 		}
 		if (ld_preload && strlen(ld_preload)) {
-			pseudo_diag("ld_preload without: <%s>\n", ld_preload);
-			setenv(PRELINK_LIBRARIES, ld_preload, 1);
+			SETENV(PRELINK_LIBRARIES, ld_preload, 1);
 		} else {
-			unsetenv(PRELINK_LIBRARIES);
+			UNSETENV(PRELINK_LIBRARIES);
 		}
 	}
 }
@@ -750,16 +898,16 @@ pseudo_setupenv() {
 	free(pseudo_get_libdir());
 	free(pseudo_get_localstatedir());
 
-        while (pseudo_env[i].key) {
+	while (pseudo_env[i].key) {
 		if (pseudo_env[i].value) {
-			setenv(pseudo_env[i].key, pseudo_env[i].value, 0);
+			SETENV(pseudo_env[i].key, pseudo_env[i].value, 0);
 			pseudo_debug(PDBGF_ENV | PDBGF_VERBOSE, "pseudo_env: %s => %s\n",
 				pseudo_env[i].key, pseudo_env[i].value);
 		}
-                i++;
-        }
+		i++;
+	}
 
-	const char *ld_library_path = getenv(PRELINK_PATH);
+	const char *ld_library_path = GETENV(PRELINK_PATH);
 	char *libdir_path = pseudo_libdir_path(NULL);
 	if (!ld_library_path) {
 		size_t len = strlen(libdir_path) + 1 + (strlen(libdir_path) + 2) + 1;
@@ -768,7 +916,7 @@ pseudo_setupenv() {
 			pseudo_diag("fatal: can't allocate new %s variable.\n", PRELINK_PATH);
 		}
 		snprintf(newenv, len, "%s:%s64", libdir_path, libdir_path);
-		setenv(PRELINK_PATH, newenv, 1);
+		SETENV(PRELINK_PATH, newenv, 1);
 	} else if (!strstr(ld_library_path, libdir_path)) {
 		size_t len = strlen(ld_library_path) + 1 + strlen(libdir_path) + 1 + (strlen(libdir_path) + 2) + 1;
 		char *newenv = malloc(len);
@@ -776,26 +924,26 @@ pseudo_setupenv() {
 			pseudo_diag("fatal: can't allocate new %s variable.\n", PRELINK_PATH);
 		}
 		snprintf(newenv, len, "%s:%s:%s64", ld_library_path, libdir_path, libdir_path);
-		setenv(PRELINK_PATH, newenv, 1);
+		SETENV(PRELINK_PATH, newenv, 1);
 	} else {
 		/* nothing to do, ld_library_path exists and contains
 		 * our preferred path */
 	}
 
-	char *ld_preload = getenv(PRELINK_LIBRARIES);
+	char *ld_preload = GETENV(PRELINK_LIBRARIES);
 	if (ld_preload) {
 		ld_preload = with_libpseudo(ld_preload, libdir_path);
 		if (!ld_preload) {
 			pseudo_diag("fatal: can't allocate new %s variable.\n", PRELINK_LIBRARIES);
 		}
-		setenv(PRELINK_LIBRARIES, ld_preload, 1);
+		SETENV(PRELINK_LIBRARIES, ld_preload, 1);
 		free(ld_preload);
 	} else {
 		ld_preload = with_libpseudo("", libdir_path);
 		if (!ld_preload) {
 			pseudo_diag("fatal: can't allocate new %s variable.\n", PRELINK_LIBRARIES);
 		}
-		setenv(PRELINK_LIBRARIES, ld_preload, 1);
+		SETENV(PRELINK_LIBRARIES, ld_preload, 1);
 		free(ld_preload);
 	}
 
@@ -806,9 +954,9 @@ pseudo_setupenv() {
 
 
 #if PSEUDO_PORT_DARWIN
-	char *force_flat = getenv("DYLD_FORCE_FLAT_NAMESPACE");
+	char *force_flat = GETENV("DYLD_FORCE_FLAT_NAMESPACE");
 	if (!force_flat) {
-		setenv("DYLD_FORCE_FLAT_NAMESPACE", "1", 1);
+		SETENV("DYLD_FORCE_FLAT_NAMESPACE", "1", 1);
 	}
 #endif
 }
@@ -847,9 +995,9 @@ pseudo_setupenvp(char * const *envp) {
 		++env_count;
 	}
 
-        for (i = 0; pseudo_env[i].key; i++) {
+	for (i = 0; pseudo_env[i].key; i++) {
 		size_pseudoenv++;
-        }
+	}
 
 	env_count += size_pseudoenv; /* We're going to over allocate */
 
@@ -1313,10 +1461,9 @@ pseudo_etc_file(const char *file, char *realname, int flags, const char **search
 }
 
 /* set up a log file */
-int
-pseudo_logfile(char *defname) {
+static int
+pseudo_logfile(char *filename, char *defname, int prefer_fd) {
 	char *pseudo_path;
-	char *filename = pseudo_get_value("PSEUDO_DEBUG_FILE");
 	char *s;
 #if PSEUDO_PORT_LINUX
 	extern char *program_invocation_short_name; /* glibcism */
@@ -1402,15 +1549,16 @@ pseudo_logfile(char *defname) {
 	if (fd == -1) {
 		pseudo_diag("help: can't open log file %s: %s\n", pseudo_path, strerror(errno));
 	} else {
-		/* try to force fd to 2.  We do this because glibc's malloc
+		/* try to force fd to prefer_fd.  We do this because glibc's malloc
 		 * debug unconditionally writes to fd 2, and we don't want
 		 * a client process ending op on fd 2, or server debugging
-		 * becomes a nightmare.
+		 * becomes a nightmare. So, server sets prefer_fd to 2. Client
+		 * leaves it at -1.
 		 */
-		if (fd != 2) {
+		if (prefer_fd >= 0 && fd != prefer_fd) {
 			int newfd;
-			close(2);
-			newfd = dup2(fd, 2);
+			close(prefer_fd);
+			newfd = dup2(fd, prefer_fd);
 			if (newfd != -1) {
 				fd = newfd;
 			}
@@ -1424,6 +1572,32 @@ pseudo_logfile(char *defname) {
 		return 0;
 }
 
+int
+pseudo_debug_logfile(char *defname, int prefer_fd) {
+	char *filename = pseudo_get_value("PSEUDO_DEBUG_FILE");
+	int fd;
+
+	fd = pseudo_logfile(filename, defname, prefer_fd);
+	if (fd > -1) {
+		pseudo_util_debug_fd = fd;
+		return 0;
+	}
+	return 1;
+}
+
+int
+pseudo_evlog_logfile(char *defname, int prefer_fd) {
+	char *filename = pseudo_get_value("PSEUDO_EVLOG_FILE");
+	int fd;
+
+	fd = pseudo_logfile(filename, defname, prefer_fd);
+	if (fd > -1) {
+		pseudo_util_evlog_fd = fd;
+		return 0;
+	}
+	return 1;
+}
+
 void
 pseudo_stat32_from64(struct stat *buf32, const struct stat64 *buf) {
 	buf32->st_dev = buf->st_dev;
